"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
import typing
import weakref
from collections import deque, defaultdict
from typing import Callable, Deque, Dict

from twisted.internet.interfaces import IPushProducer
from zope.interface import implementer

from vortex.DeferUtil import nonConcurrentMethod

logger = logging.getLogger(name=__name__)


def _format_size(size):
    for unit in ('B', 'KiB', 'MiB', 'GiB', 'TiB'):
        if abs(size) < 100 and unit != 'B':
            # 3 digits (xx.x UNIT)
            return "%.1f %s" % (size, unit)
        if abs(size) < 10 * 1024 or unit == 'TiB':
            # 4 or 5 digits (xxxx UNIT)
            return "%.0f %s" % (size, unit)
        size /= 1024


@implementer(IPushProducer)
class VortexWritePushProducer(object):
    WARNING_DATA_LENGTH = 50 * 1024 * 1024
    ERROR_DATA_LENGTH = 500 * 1024 * 1024
    WRITE_CHUNK_SIZE = 128 * 1024

    __memoryLoggingRefs = None
    __memoryLoggingEnabled = False

    @classmethod
    def setupMemoryLogging(cls) -> None:
        cls.__memoryLoggingRefs = []
        cls.__memoryLoggingEnabled = True

    @classmethod
    def memoryLoggingDump(cls, top=10,
                          msgs=1) -> typing.List[typing.Tuple[str, int, int]]:
        if not cls.__memoryLoggingRefs:
            return []

        # Filter out expired items
        cls.__memoryLoggingRefs = list(filter(lambda o: o(), cls.__memoryLoggingRefs))

        results = []
        for producerRef in cls.__memoryLoggingRefs:
            producer = producerRef()
            if not producer:
                continue

            queueCount = sum([len(q) for q in producer._queueByPriority.values()])

            results.append((producer._remoteVortexName,
                            queueCount,
                            producer._queuedDataLen))

        data = sorted(results, key=lambda x: x[2], reverse=True)

        return list(filter(lambda x: x[2] >= msgs, data))[:top]

    def __init__(self, transport,
                 stopProducingCallback: Callable,
                 remoteVortexName: str = 'Pending',
                 writeWholeFrames=False):
        if VortexWritePushProducer.__memoryLoggingEnabled:
            VortexWritePushProducer.__memoryLoggingRefs.append(weakref.ref(self))

        self._transport = transport
        self._stopProducingCallback = stopProducingCallback
        self._remoteVortexName = remoteVortexName
        self._writeWholeFrames = writeWholeFrames

        self._paused = False
        self._pausedDeferred = None
        self._writingInProgress = False
        self._writingFrameInProgress = False
        self._closed = False
        self._queuedDataLen = 0
        self._queueByPriority: Dict[int, Deque] = defaultdict(deque)

        self._currentlyWritingData: typing.Optional[bytes] = None
        self._currentlyWritingDataOffset: int = 0

    def setRemoteVortexName(self, remoteVortexName: str):
        self._remoteVortexName = remoteVortexName

    @property
    def _canContinue(self):
        return not self._paused \
               and not self._startWritingFrame.running \
               and [q for q in self._queueByPriority.values() if q]

    @nonConcurrentMethod
    def _startWriting(self):
        if self._currentlyWritingData is not None:
            self._startWritingFrame()

        while self._canContinue:
            # Send the messages in order of priority
            for priority in sorted(self._queueByPriority):
                if not self._canContinue:
                    return

                queue = self._queueByPriority[priority]
                if not queue:
                    continue

                data = queue.popleft()
                preLen = self._queuedDataLen
                self._queuedDataLen -= len(data)

                if self._queuedDataLen < self.WARNING_DATA_LENGTH < preLen:
                    logger.info(
                        "%s: Data Queue memory high warning - returned to normal : %s",
                        self._remoteVortexName,
                        _format_size(self._queuedDataLen))

                if self._queuedDataLen < self.ERROR_DATA_LENGTH < preLen:
                    logger.info(
                        "%s: Data Queue memory high error - returned to warning : %s",
                        self._remoteVortexName,
                        _format_size(self._queuedDataLen))

                if self._writeWholeFrames:
                    self._transport.write(data)
                    self._transport.write(b'.')

                else:
                    self._currentlyWritingData = data
                    self._currentlyWritingDataOffset = 0
                    self._startWritingFrame()

    @nonConcurrentMethod
    def _startWritingFrame(self):
        # ---------------
        # Write in progress logic.
        # We should only have one write loop at a time
        if self._writingFrameInProgress:
            return
        self._writingFrameInProgress = True

        offset = self._currentlyWritingDataOffset
        data = self._currentlyWritingData
        while not self._paused and offset < len(data):
            self._transport.write(data[offset:offset + self.WRITE_CHUNK_SIZE])
            offset += self.WRITE_CHUNK_SIZE

        if len(data) <= offset:
            self._currentlyWritingDataOffset = 0
            self._currentlyWritingData = None
            self._transport.write(b'.')

        else:
            self._currentlyWritingDataOffset = offset

        self._writingFrameInProgress = False

    def pauseProducing(self):
        """
        Pause producing data.

        Tells a producer that it has produced too much data to process for
        the time being, and to stop until C{resumeProducing()} is called.
        """
        self._paused = True
        # logger.debug("%s: Producer paused, data len = %s",
        #              self._remoteVortexName, self._queuedDataLen)

    def resumeProducing(self):
        """
        Resume producing data.

        This tells a producer to re-add itself to the main loop and produce
        more data for its consumer.
        """
        self._paused = False
        # logger.debug("%s: Producer resumed, data len = %s",
        #              self._remoteVortexName, self._queuedDataLen)
        self._startWriting()

    def stopProducing(self):
        """
        Stop producing data.

        This tells a producer that its consumer has died, so it must stop
        producing data for good.
        """
        self._stopProducingCallback()

    def write(self, data, priority: int):
        assert not self._closed

        preLen = self._queuedDataLen
        self._queuedDataLen += len(data)

        # Queue the data up in chunks
        self._queueByPriority[priority].append(data)

        if preLen < self.WARNING_DATA_LENGTH < self._queuedDataLen:
            logger.warning("%s: Data Queue memory high warning : %s",
                           self._remoteVortexName,
                           _format_size(self._queuedDataLen))

        if preLen < self.ERROR_DATA_LENGTH < self._queuedDataLen:
            logger.error("%s: Data Queue memory high error : %s",
                         self._remoteVortexName,
                         _format_size(self._queuedDataLen))

        self._startWriting()

    def close(self):
        self._closed = True



import unittest

from twisted.internet import reactor
from twisted.internet.defer import DeferredList
from twisted.internet.endpoints import TCP4ServerEndpoint
from twisted.internet.protocol import Factory, Protocol, connectionDone

from vortex.Payload import Payload
from vortex.VortexClientHttp import VortexClientHttp

class DummyListenerProtocol(Protocol):

    def connectionMade(self):
        pass

    def dataReceived(self, data):
        pass

    def connectionLost(self, reason=connectionDone):
        pass


class DummyListenerFactory(Factory):

    def buildProtocol(self, addr):
        return DummyListenerProtocol()



class VortexClientTest(unittest.TestCase):
    def setUp(self):
        port = 9009
        host = "localhost"

        def listenSuccess(listeningPort):
            self.listeningPort = listeningPort

            self.vortexClient = VortexClientHttp()
            return self.vortexClient.connect(host, port)

        endpoint = TCP4ServerEndpoint(reactor, port)
        d = endpoint.listen(DummyListenerFactory())
        d.addCallback(listenSuccess)

        return d

    def tearDown(self):
        try:
            self.vortexClient.close()
        except:
            pass

        self.listeningPort.stopListening()

    def testSend(self):
        deferreds = []

        # Just test functionality.
        d = self.vortexClient.send(Payload())
        deferreds.append(d)

        d = self.vortexClient.send([Payload(), Payload()])
        deferreds.append(d)

        d = self.vortexClient.sendVortexMsg(Payload().toVortexMsg())
        deferreds.append(d)

        d = self.vortexClient.sendVortexMsg([Payload().toVortexMsg(), Payload().toVortexMsg()])
        deferreds.append(d)

        return DeferredList(deferreds)




"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
import traceback
import typing
from copy import copy, deepcopy
from json import JSONEncoder
from typing import List
from _collections import defaultdict

logger = logging.getLogger(__name__)

from .SerialiseUtil import *

###############################################################################
# JSON Constants
###############################################################################

# Globally mess with the json dump separators for more compact json
JSONEncoder.item_separator = ','
JSONEncoder.key_separator = ':'


class Jsonable(object):
    ''' Xmlable
    Inherit from this class if you want to serialise to XML.
    '''
    __fieldNames__: List[str] = None
    __rapuiSerialiseType__ = T_GENERIC_CLASS

    __rawJonableFields__: List[str] = None

    JSON_CLASS_TYPE = '_ct'
    JSON_CLASS = '_c'
    JSON_TUPLE_TYPE = '_c'
    JSON_FIELD_TYPE = "_ft"
    JSON_FIELD_DATA = "_fd"

    __memoryLoggingRefs = None
    __memoryLoggingEnabled = False

    @classmethod
    def setupMemoryLogging(cls) -> None:
        cls.__memoryLoggingRefs = defaultdict(lambda: 0)
        cls.__memoryLoggingEnabled = True

    @classmethod
    def memoryLoggingDump(cls, top=10, over=100) -> List[typing.Tuple[str, int]]:
        if not Jsonable.__memoryLoggingRefs:
            return []

        data = sorted(Jsonable.__memoryLoggingRefs.items(),
                      key=lambda x: x[1],
                      reverse=True)

        return list(filter(lambda x: x[1] >= over, data))[:top]

    def __memLoggingKey(self):
        if getattr(self, '__rapuiSerialiseType__') == T_RAPUI_TUPLE:
            key = 'Tuple: ' + getattr(self, '__tupleType__')

        elif getattr(self, '__rapuiSerialiseType__') == T_RAPUI_PAYLOAD:
            assert hasattr(self, 'filt'), 'Payload is missing filt'
            if 'key' in self.filt:
                key = 'Payload: key=' + self.filt['key']
            else:
                key = 'Payload: ' + str(self.filt)

        else:
            key = 'Jsonable: ' + getattr(self, '__rapuiSerialiseType__')

        return key

    def __init__(self):
        if Jsonable.__memoryLoggingEnabled:
            Jsonable.__memoryLoggingRefs[self.__memLoggingKey()] += 1

    def __del__(self):
        if Jsonable.__memoryLoggingEnabled:
            Jsonable.__memoryLoggingRefs[self.__memLoggingKey()] -= 1

    def __isRawJsonableField(self, name: str) -> bool:
        if not (name and self.__rawJonableFields__):
            return False
        return name in self.__rawJonableFields__

    def toJsonField(self, value, jsonDict=None, name=None):
        ''' To Json Field

        '''
        from .Tuple import TupleField
        if isinstance(value, TupleField):
            value = None if value.defaultValue == None else copy(value.defaultValue)

        try:
            valueType = V_NULL if value is None else toRapuiType(value)
        except KeyError as e:
            raise KeyError("%s field name is %s, type is %s",
                           e, name, value.__class__)

        # Payloads and Tuples
        if self.__isRawJsonableField(name):
            # This sho
            convertedValue = deepcopy(value)

        elif isinstance(value, Jsonable):
            convertedValue = value.toJsonDict()

        elif isinstance(value, dict):
            convertedValue = {}
            for key, value in list(value.items()):
                self.toJsonField(value, convertedValue, key)

        elif isinstance(value, (set, list, tuple)):
            convertedValue = []
            for val in value:
                convertedValue.append(self.toJsonField(val))

        # Decimals are detected as floats for some reason, so convert it here
        elif valueType == T_FLOAT:
            convertedValue = float(value)

        elif valueType in (T_INT, T_BOOL, T_STR):
            convertedValue = value

        elif valueType == V_NULL:
            convertedValue = None

        else:
            convertedValue = toStr(value)

        # Non standard values need a dict to store their value type attributes
        # Create a sub dict that contains the value and type
        if (valueType not in (T_FLOAT, T_STR, V_NULL, T_BOOL, T_LIST, T_DICT)
                and not isinstance(value, Jsonable)):
            convertedValue = {
                Jsonable.JSON_FIELD_TYPE: valueType,
                Jsonable.JSON_FIELD_DATA: convertedValue
            }

        if name and jsonDict is not None:
            # Now assign the value and it's data type if applicable
            if not toRapuiType(name) in (T_STR, T_INT, T_FLOAT):
                raise Exception("name=%s, type=%s, is not an allowed dict key",
                                name, toRapuiType(name))

            jsonDict[name] = convertedValue

        return convertedValue

    # -----------------------------------------------------------------------------
    def fromJsonField(self, value, valueType=None):
        # Single Value
        if valueType == V_NULL or value is None:  # V_NULL will never be set in toJsonField
            return None

        if valueType == T_INT:
            return int(value)

        if isinstance(value, dict) and Jsonable.JSON_CLASS_TYPE in value:
            valueType = value[Jsonable.JSON_CLASS_TYPE]

        # JSON handles these types natively,
        # if there is no type then these are the right types
        if valueType is None:
            valueType = toRapuiType(value)
            if valueType in (T_BOOL, T_FLOAT, T_INT, T_STR):
                return value

        # Non standard values need a dict to store their value type attributes, decode these
        if isinstance(value, dict) and Jsonable.JSON_FIELD_TYPE in value:
            return self.fromJsonField(value[Jsonable.JSON_FIELD_DATA],
                                      value[Jsonable.JSON_FIELD_TYPE])

        # Tuple
        if valueType == T_RAPUI_TUPLE:
            tupleType = value[Jsonable.JSON_TUPLE_TYPE]

            from .Tuple import TUPLE_TYPES_BY_NAME
            if not tupleType in TUPLE_TYPES_BY_NAME:
                raise Exception("Tuple type |%s| not registered within this program.",
                                tupleType)

            try:
                from .Tuple import TUPLE_TYPES_BY_NAME
                return TUPLE_TYPES_BY_NAME[tupleType]().fromJsonDict(value)

            except Exception as e:
                logger.critical(traceback.format_exc())
                raise Exception("%s for tuple type %s" % (str(e), tupleType))

        # Handle the case of payloads within payloads
        if valueType == T_RAPUI_PAYLOAD:
            from .Payload import Payload
            return Payload().fromJsonDict(value)

        if valueType == T_RAPUI_PAYLOAD_ENVELOPE:
            from .PayloadEnvelope import PayloadEnvelope
            return PayloadEnvelope().fromJsonDict(value)

        if valueType == T_GENERIC_CLASS:
            # OTHER JSONABLES GO HERE, INSPECT THE tuple_TYPE
            # jsonClass = value[JSON_CLASS]
            raise NotImplementedError()
            # import importlib
            # MyClass = getattr(importlib.import_module("module.submodule"), "Klass")
            # instance = MyClass()

        if valueType == T_DICT:
            # Dict
            restoredDict = {}
            for subName, subVal in list(value.items()):
                try:
                    subNamStr = subName
                    subName = float(subNamStr)
                    subName = int(subNamStr)
                except ValueError:
                    pass
                restoredDict[subName] = self.fromJsonField(subVal)

            return restoredDict

        if valueType == T_LIST:
            restoredList = []
            for subVal in value:
                restoredList.append(self.fromJsonField(subVal))

            return restoredList

        return fromStr(value, valueType)

    def toJsonDict(self):

        fieldNames = self.__fieldNames__

        jsonDict = {Jsonable.JSON_CLASS_TYPE: self.__rapuiSerialiseType__}

        if hasattr(self, 'tupleName'):
            jsonDict[Jsonable.JSON_TUPLE_TYPE] = self.tupleName()

        else:
            jsonDict[Jsonable.JSON_CLASS] = className(self)

        for name in fieldNames:
            data = getattr(self, name)
            self.toJsonField(data, jsonDict, name)

        return jsonDict

    def fromJsonDict(self, jsonDict):
        ''' From Xml
        Returns and instance of this object populated with data from the json dict
        '''

        # Use the fromJsonField code to convert all the values
        for name, value in list(jsonDict.items()):
            if name.startswith("_"):
                continue

            if self.__isRawJsonableField(name):
                setattr(self, name, value)
            else:
                setattr(self, name, self.fromJsonField(value))

        return self



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""

import collections
import logging
from _pyio import BytesIO
from typing import Deque

from twisted.internet import task
from twisted.internet.defer import inlineCallbacks
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET

from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexServer import HEART_BEAT_PERIOD
from .PayloadPriority import DEFAULT_PRIORITY
from .VortexConnectionABC import VortexConnectionABC
from .VortexServer import VortexServer

logger = logging.getLogger(name=__name__)


class VortexServerHttpResource(Resource):
    """ VortexServer Resource
      This resource is the server endpoint for the vortex.
    """
    isLeaf = True

    def __init__(self, vortex: VortexServer) -> None:
        Resource.__init__(self)
        self.__vortex = vortex

    def render_GET(self, request):
        request.code = 405
        return b"You have reached the vortex. It only likes POST methods."

    def render_POST(self, request):
        remoteVortexUuid = request.args[b'vortexUuid'][0].decode()
        remoteVortexName = request.args[b'vortexName'][0].decode()

        if self.__vortex.isShutdown():
            return None

        httpSession = request.getSession()
        conn = VortexResourceConnection(self.__vortex,
                                        remoteVortexUuid,
                                        remoteVortexName,
                                        request)

        # Send a heart beat down the new connection, tell it who we are.
        connectPayloadFilt = {}
        connectPayloadFilt[PayloadEnvelope.vortexUuidKey] = self.__vortex.uuid()
        connectPayloadFilt[PayloadEnvelope.vortexNameKey] = self.__vortex.name()
        conn.write(PayloadEnvelope(filt=connectPayloadFilt).toVortexMsg())

        data = request.content.read()
        if len(data):
            for vortexStr in data.strip(b'.').split(b'.'):
                self._processVortexMsg(httpSession, conn, vortexStr.decode("UTF-8"))

        # Request will be around for a while, do some cleanups
        request.content = BytesIO()
        request.args = {}

        self.__vortex.connectionOpened(httpSession, conn)

        def connClosed(err):
            logger.debug("VortexServer connection ended by client")
            self.__vortex.connectionClosed(conn)

        request.notifyFinish().addErrback(connClosed)

        return NOT_DONE_YET

    @inlineCallbacks
    def _processVortexMsg(self, httpSession, conn, vortexMsg):
        payload = yield PayloadEnvelope().fromVortexMsgDefer(vortexMsg)
        self.__vortex.payloadReveived(
            httpSession=httpSession,
            vortexUuid=conn.remoteVortexUuid,
            vortexName=conn.remoteVortexName,
            payload=payload)


class VortexResourceConnection(VortexConnectionABC):
    def __init__(self, vortexServer: VortexServer,
                 remoteVortexUuid: str,
                 remoteVortexName: str,
                 request) -> None:
        VortexConnectionABC.__init__(self,
                                     logger,
                                     vortexServer,
                                     remoteVortexUuid=remoteVortexUuid,
                                     remoteVortexName=remoteVortexName,
                                     httpSessionUuid=request.getSession().uid)

        self._request = request
        self._request.responseHeaders.setRawHeaders(b'content-type'
                                                    , [b'text/text'])

        self._buffer: Deque[bytes] = collections.deque()

        self._requestReadyCheckTimer = task.LoopingCall(self._checkIfRequestIsReady)
        self._requestReadyCheckTimer.start(0.001)

        # Start our heart beat
        self._beatLoopingCall = task.LoopingCall(self._beat)
        d = self._beatLoopingCall.start(HEART_BEAT_PERIOD)
        d.addErrback(lambda f: logger.exception(f.value))

    def _beat(self):
        if self._closed or self._request.finished or self._request.channel is None:
            self._beatLoopingCall.stop()
            return

        # Send the heartbeats
        self._request.write(b'.')

        # Touch the session
        self._request.getSession().touch()

    @property
    def ip(self):
        return self._request.client.host

    @property
    def port(self):
        return self._request.client.port

    def _checkIfRequestIsReady(self):
        # This seems to be the fingerprint of the fact that request will not write
        # our data yet.
        if self._request.chunked and not self._request.startedWriting:
            return

        if self._requestReadyCheckTimer.running:
            self._requestReadyCheckTimer.stop()

        while self._buffer:
            self._write(self._buffer.popleft())
        self._buffer = None

    def write(self, payloadVortexStr: bytes, priority: int = DEFAULT_PRIORITY):
        assert not self._closed
        # If we have a buffer, write it.
        if self._buffer:
            self._buffer.append(payloadVortexStr)
        else:
            self._write(payloadVortexStr)

    def _write(self, payloadVortexStr: bytes):
        if self._request.finished or self._request.channel is None:
            logger.error("VORTEX CLOSED, CAN'T WRITE PAYLOAD")
            self._closed = True
            self._vortexServer.connectionClosed(self)
            return

        self._request.write(payloadVortexStr)
        self._request.write(b'.')

    def close(self):
        if self._beatLoopingCall.running:
            self._beatLoopingCall.stop()

        self._closed = True
        self._request.finish()



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""

from twisted.trial import unittest

from run_test_vortex_server import setupVortexServer
from vortex.VortexClientHttp import VortexClientHttp


class VortexTest(unittest.TestCase):
    def setUp(self):
        self.port = setupVortexServer()

    def testVortexConnect(self):
        client = VortexClientHttp("unittest")
        return client.connect("127.0.0.1", self.port)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import decimal
import logging
from datetime import datetime, date

from base64 import b64encode, b64decode
from collections import defaultdict
from sqlalchemy.orm.collections import InstrumentedList

logger = logging.getLogger(__name__)

try:
    from geoalchemy2.elements import WKBElement

except ImportError:
    class WKBElement:
        def __init__(self):
            raise NotImplementedError("")

try:
    from geoalchemy2.elements import WKBElement

except ImportError:
    class WKBElement:
        def __init__(self):
            raise NotImplementedError("")

###############################################################################
# Utility functions
###############################################################################

T_RAPUI_TUPLE = 'rt'
T_RAPUI_PAYLOAD = 'rp'
T_RAPUI_PAYLOAD_ENVELOPE = 'rpe'
T_GENERIC_CLASS = 'gen'  # NOT SUPPORTED
T_FLOAT = 'float'
T_INT = 'int'
T_STR = 'str'
T_BYTES = 'bytes'
T_BOOL = 'bool'
T_DATETIME = 'datetime'
T_DICT = 'dict'
T_LIST = 'list'
T_WKB = 'wkb' # Well Known Binary, a part of the GEO library

V_NULL = 'null'
V_TRUE = '1'
V_FALSE = '0'

ISO8601_ORA = 'YYYY-MM-DD HH24:MI:SS.FF6'
ISO8601 = '%Y-%m-%d %H:%M:%S.%f%z'
ISO8601_REGEXP = r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}[+-]\d{2}:\d{2}$'
ISO8601_NOTZ = '%Y-%m-%d %H:%M:%S.%f'

TYPE_MAP_PY_TO_RAPUI = {decimal.Decimal: T_FLOAT,
                        float: T_FLOAT,
                        str: T_STR,
                        bytes: T_BYTES,
                        int: T_INT,
                        bool: T_BOOL,
                        datetime: T_DATETIME,
                        date: T_DATETIME,
                        dict: T_DICT,
                        defaultdict: T_DICT,
                        list: T_LIST,
                        set: T_LIST,
                        tuple: T_LIST,
                        WKBElement: T_WKB,
                        InstrumentedList: T_LIST
                        }


def className(o):
    cls = o if o.__class__ == type else o.__class__
    return str(cls).split("'")[1]


def convertFromShape(shapelyShape):
    from shapely.geometry.polygon import Polygon
    if isinstance(shapelyShape, Polygon):
        coords = shapelyShape.exterior.coords
    else:
        coords = shapelyShape.coords
    return [{'x': i[0], 'y': i[1]} for i in coords]


def convertFromWkbElement(value):
    from geoalchemy2.shape import to_shape
    return convertFromShape(to_shape(value))


def serialiseWkb(wkb:WKBElement) -> str:
    from geoalchemy2.shape import to_shape
    import shapely
    return b64encode(shapely.wkb.dumps(to_shape(wkb))).decode()

def deserialiseWkb(wkbB64Data:str) -> WKBElement:
    from geoalchemy2.shape import from_shape
    import shapely
    return from_shape(shapely.wkb.loads(b64decode(wkbB64Data)))


TYPE_MAP_RAPUI_TO_PY = {T_FLOAT: float,
                        T_STR: str,
                        T_BYTES: bytes,
                        T_INT: int,
                        T_BOOL: bool,
                        T_DATETIME: datetime,
                        T_DICT: dict,
                        T_LIST: list,
                        T_WKB: deserialiseWkb
                        }


def decimalToStr(dec):
    """Return string representation of the number in scientific notation.

    Captures all of the information in the underlying representation.
    """

    if str(dec) == '0E-10':
        return '0'

    sign = ['', '-'][dec._sign]
    if dec._is_special:
        if dec._exp == 'F':
            return sign + 'Infinity'
        elif dec._exp == 'n':
            return sign + 'NaN' + dec._int
        else:  # dec._exp == 'N'
            return sign + 'sNaN' + dec._int

    # number of digits of dec._int to left of decimal point
    leftdigits = dec._exp + len(dec._int)

    # no exponent required
    dotplace = leftdigits

    if dotplace <= 0:
        intpart = '0'
        fracpart = '.' + '0' * (-dotplace) + dec._int
    elif dotplace >= len(dec._int):
        intpart = dec._int + '0' * (dotplace - len(dec._int))
        fracpart = ''
    else:
        intpart = dec._int[:dotplace]
        fracpart = '.' + dec._int[dotplace:]

    if leftdigits == dotplace:
        exp = ''
    else:
        exp = "E%+d" % (leftdigits - dotplace)

    s = (sign + intpart + fracpart + exp)
    if '.' in s:
        s = s.rstrip('0').rstrip('.')
    assert (len(s))
    return s


def toStr(obj) -> str:
    if isinstance(obj, WKBElement):
        # noinspection PyTypeChecker
        return serialiseWkb(obj)

    if isinstance(obj, decimal.Decimal):
        return decimalToStr(obj)

    if isinstance(obj, datetime):
        return obj.strftime(ISO8601)

    if isinstance(obj, bool):
        return V_TRUE if obj else V_FALSE

    if isinstance(obj, bytes):
        return b64encode(obj).decode() # to make it a str

    if isinstance(obj, str):
        return obj

    try:
        return str(obj)
    except:
        logger.critical("Failed to convert the following to string")
        logger.critical(obj)
        raise


def fromStr(val: str, typeName: str):
    if typeName == T_DATETIME:
        if len(val) <= len('2017-06-06 09:40:37.097068'):
            return datetime.strptime(val, ISO8601_NOTZ)
        return datetime.strptime(val, ISO8601)

    if typeName == T_BOOL:
        return val == V_TRUE

    if typeName == T_BYTES:
        return b64decode(val)

    if typeName in (T_STR):
        return val

    pyVal = fromRapuiType(typeName)(val)

    if typeName == T_FLOAT and pyVal.is_integer():
        pyVal = int(pyVal)

    return pyVal


def fromRapuiType(typeName: str):
    return TYPE_MAP_RAPUI_TO_PY[typeName]


def toRapuiType(value):
    if value is None:
        return V_NULL

    from .Tuple import Tuple
    if isinstance(value, Tuple):
        return T_RAPUI_TUPLE

    from .Payload import Payload
    if isinstance(value, Payload):
        return T_RAPUI_PAYLOAD

    from .PayloadEnvelope import PayloadEnvelope
    if isinstance(value, PayloadEnvelope):
        return T_RAPUI_PAYLOAD_ENVELOPE

    valueClass = value.__class__

    return TYPE_MAP_PY_TO_RAPUI[valueClass]



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
from datetime import datetime

import pytz
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure

from vortex.DeferUtil import vortexLogFailure
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexABC import SendVortexMsgResponseCallable

logger = logging.getLogger(name="PayloadIO")


class PayloadIO(object):
    '''
    PayloadIO, Processes payloads received from the vortex and distributes
    them to where they need to go.
    '''

    # Singleton
    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(PayloadIO, cls).__new__(cls)
            cls._instance.__singleton_init__()
        return cls._instance

    def __singleton_init__(self):
        self._endpoints = set()

    def remove(self, endpoint):
        if endpoint in self._endpoints:
            self._endpoints.remove(endpoint)

    def add(self, endpoint):
        self._endpoints.add(endpoint)

    @property
    def endpoints(self):
        ''' Endpoints property

        @:return a copy of the list of endpoints
        '''
        return list(self._endpoints)

    def process(self, payloadEnvelope: PayloadEnvelope,
                vortexUuid: str, vortexName: str, httpSession,
                sendResponse: SendVortexMsgResponseCallable):

        immutableEndpoints = list(self._endpoints)
        for endpoint in immutableEndpoints:
            if not endpoint.check(payloadEnvelope, vortexName):
                continue

            reactor.callLater(0, self._processLater, endpoint, payloadEnvelope,
                              vortexUuid, vortexName, httpSession, sendResponse)

    def _processLater(self, endpoint,
                      payloadEnvelope: PayloadEnvelope,
                      vortexUuid: str, vortexName: str, httpSession,
                      sendResponse: SendVortexMsgResponseCallable):
        startDate = datetime.now(pytz.utc)

        def respondToException(failure):
            """ Respond To Exception
            Putting the exception into a failure messes with the stack, hence the
            common function
            """
            try:
                sendResponse(
                    PayloadEnvelope(filt=payloadEnvelope.filt, result=str(failure.getTraceback()))
                    .toVortexMsg()
                )
            except Exception as e:
                logger.exception(e)

            vortexLogFailure(failure, logger)
            logger.error(payloadEnvelope.filt)

        def callback(value, blocking=False):
            # A blocking call taking more than 100ms is BAD
            # Otherwise a call taking more than a 1s is just poor performance.
            secondsTaken = (datetime.now(pytz.utc) - startDate).total_seconds()
            if secondsTaken > 0.1 and blocking or secondsTaken > 2.0:
                func = logger.critical if blocking else logger.debug
                func("%s Payload endpoint took %s\npayload.filt=%s\n%s",
                     'BLOCKING ' if blocking else '',
                     secondsTaken,
                     payloadEnvelope.filt,
                     endpoint)

        try:
            d = endpoint.process(payloadEnvelope,
                                 vortexUuid, vortexName, httpSession, sendResponse)
            if isinstance(d, Deferred):
                d.addCallback(callback)
                d.addErrback(respondToException)
            elif hasattr(d, 'addCallback'):
                raise Exception("isinstance FAILED")
            else:
                callback(True, blocking=True)

        except Exception as e:
            respondToException(Failure(e))



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
from urllib.parse import urlparse, parse_qs

from twisted.internet.defer import inlineCallbacks
from twisted.internet.protocol import Protocol, connectionDone, Factory
from twisted.protocols.policies import WrappingFactory
from twisted.web import resource
from twisted.web.server import NOT_DONE_YET
from txwebsocket.txws import WebSocketProtocol

from vortex.DeferUtil import vortexLogFailure
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexServerConnection import VortexServerConnection
from .PayloadPriority import DEFAULT_PRIORITY
from .VortexServer import VortexServer

logger = logging.getLogger(name=__name__)


class VortexWebsocketServerProtocol(Protocol):
    def __init__(self, vortex: VortexServer, addr, httpSession=None) -> None:
        self._vortex = vortex
        self._addr = addr

        self._conn = None

        self._remoteVortexUuid = None
        self._remoteVortexName = None
        self._httpSession = httpSession

        def httpSessionExpiredCallback():
            self._httpSession = None
            self.transport.loseConnection()

        if self._httpSession:
            self._httpSession.notifyOnExpire(httpSessionExpiredCallback)

        # Most messages don't need a buffer, but websockets can split messages
        # around the 128kb mark
        self._receiveBuffer = b''

    def __initConnection(self):
        # self.transport.setBinaryMode(True)

        params = parse_qs(urlparse(self.transport.location).query)

        if 'vortexUuid' not in params or 'vortexName' not in params:
            raise Exception("This isn't a vortex capable websocket. Check the URL")

        self._remoteVortexUuid = params['vortexUuid'][0]
        self._remoteVortexName = params['vortexName'][0]
        self._conn = VortexServerConnection(self._vortex,
                                            self._remoteVortexUuid,
                                            self._remoteVortexName,
                                            self._httpSession,
                                            self.transport,
                                            self._addr)

        # Send a heart beat down the new connection, tell it who we are.
        connectPayloadFilt = {}
        connectPayloadFilt[PayloadEnvelope.vortexUuidKey] = self._vortex.uuid()
        connectPayloadFilt[PayloadEnvelope.vortexNameKey] = self._vortex.name()
        self._conn.write(PayloadEnvelope(filt=connectPayloadFilt).toVortexMsg(),
                         DEFAULT_PRIORITY)

        self._vortex.connectionOpened(self._httpSession, self._conn)

    def dataReceived(self, data):

        if not self._conn:
            self.__initConnection()

        if self._vortex.isShutdown():
            return None

        if self._httpSession:
            self._httpSession.touch()

        if data == b'.':
            self._conn.beatReceived()
            return

        self._receiveBuffer += data

        if self._receiveBuffer.endswith(b'.'):
            d = self._processVortexMsg(self._receiveBuffer)
            d.addErrback(vortexLogFailure, logger, consumeError=True)
            self._receiveBuffer = b''

    def connectionLost(self, reason=connectionDone):
        if self._conn:
            self._conn.transportClosed()

    @inlineCallbacks
    def _processVortexMsg(self, vortexMsg: bytes):
        payloadEnvelope = yield PayloadEnvelope().fromVortexMsgDefer(vortexMsg)
        self._vortex.payloadReveived(
            httpSession=self._httpSession,
            vortexUuid=self._remoteVortexUuid,
            vortexName=self._remoteVortexName,
            payload=payloadEnvelope)


class VortexWebsocketServerFactory(Factory):
    protocol = None

    def __init__(self, vortexServer: VortexServer) -> None:
        self._vortexServer = vortexServer

    def buildProtocol(self, addr, httpSession=None):
        p = VortexWebsocketServerProtocol(self._vortexServer, addr, httpSession)
        p.factory = self
        return p


class VortexWrappedWebSocketFactory(WrappingFactory):
    """
    Factory which wraps another factory to provide WebSockets transports for
    all of its protocols.
    """

    protocol = WebSocketProtocol

    def buildProtocol(self, addr, httpSession):
        return self.protocol(self, self.wrappedFactory.buildProtocol(addr, httpSession))


class VortexWebSocketUpgradeResource(resource.Resource):
    """ Vortex Websocket Upgrade Resource

    If this resource is hit, it will attempt to upgrade the connection to a websocket.

    """
    isLeaf = 1

    def __init__(self, websocketFactory):
        """ Constructor

        @:param websocketFactory: A factory that will build a WebsocketProtocol (above)
        """
        resource.Resource.__init__(self)
        self._websocketFactory = websocketFactory

    def render(self, request):
        httpSession = request.getSession()

        websocketProtocol = self._websocketFactory \
            .buildProtocol(request.client.host, httpSession)
        websocketProtocol.makeConnection(request.channel.transport)
        websocketProtocol.initFromRequest(request)
        request.channel.upgradeToWebsocket(websocketProtocol)

        return NOT_DONE_YET



import logging

from twisted.internet import defer
from twisted.internet.defer import Deferred

from vortex import TupleAction
from vortex.TupleAction import TupleActionABC, TupleGenericAction
from vortex.handler.TupleActionProcessor import TupleActionProcessor, \
    TupleActionProcessorDelegateABC
from vortex.test.PerformTestActionTuple import PerformTestActionTuple

logger = logging.getLogger(__name__)


class TestProcessor(TupleActionProcessorDelegateABC):
    def processTupleAction(self, tupleAction: TupleActionABC) -> Deferred:
        assert isinstance(tupleAction, TupleGenericAction), (
            "tupleAction is not TupleGenericAction"
        )
        if "FAIL PLEASE" == tupleAction.data:
            raise Exception("Processing failed scenario, exception raised")
        return defer.succeed(True)


actionProcessor = TupleActionProcessor("vortexTestActions")
actionProcessor.setDelegate(PerformTestActionTuple.tupleName(), TestProcessor())



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
from .Tuple import Tuple, TupleField, addTupleType


@addTupleType
class DataWrapTuple(Tuple):
    ''' Data Wrap Tuple

    This tuple wraps a peice of data for use in RapUI.

    Attributes:
      name         What ever you want to use it for
      data         The data

    '''

    __tupleType__ = 'c.s.r.datawraptuple'

    name = TupleField(None)
    data = TupleField(None)



import logging
from copy import copy

from twisted.internet.defer import TimeoutError, inlineCallbacks
from twisted.python.failure import Failure

from vortex.DeferUtil import deferToThreadWrapWithLogger, vortexLogFailure
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.PayloadResponse import PayloadResponse
from vortex.TupleSelector import TupleSelector
from vortex.VortexABC import SendVortexMsgResponseCallable
from vortex.VortexFactory import VortexFactory
from vortex.handler.TupleDataObservableCache import TupleDataObservableCache
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler, \
    TuplesProviderABC

logger = logging.getLogger(__name__)


class TupleDataObservableProxyHandler(TupleDataObservableCache):
    __CHECK_PERIOD = 30  # seconds

    def __init__(self, observableName, proxyToVortexName: str,
                 additionalFilt=None, subscriptionsEnabled=True,
                 observerName="default") -> None:
        """ Constructor

        :param observableName: The name of this and the other observable
        :param proxyToVortexName: The vortex dest name to proxy requests to
        :param additionalFilt: Any additional filter keys that are required
        :param subscriptionsEnabled: Should subscriptions be enabled (default)
        :param observerName: We can clash with other observers, so where there are
        multiple observers on the one vortex, they should use different names.
        """
        TupleDataObservableCache.__init__(self)

        self._proxyToVortexName = proxyToVortexName
        self._subscriptionsEnabled = subscriptionsEnabled
        self._observerName = observerName
        self._filt = dict(name=observableName,
                          key="tupleDataObservable")
        if additionalFilt:
            self._filt.update(additionalFilt)

        # Create the local observable, this allows local tuple providers
        # The rest are proxied on to the backend
        self._localObservableHandler = TupleDataObservableHandler(
            observableName,
            additionalFilt=additionalFilt,
            subscriptionsEnabled=subscriptionsEnabled)
        # Shutdown the local observables endpoint, we don't want it listening it's self
        self._localObservableHandler.shutdown()

        # Finally, Setup our endpoint
        self._endpoint = PayloadEndpoint(self._filt, self._process)

        TupleDataObservableCache.start(self)

    def shutdown(self):
        self._endpoint.shutdown()
        TupleDataObservableCache.shutdown(self)

    ## ----- Implement local observable

    def addTupleProvider(self, tupleName, provider: TuplesProviderABC):
        """ Add Tuple Provider

        Adds a tuple provider to the local observable.

        All other requests are proxied on

        """
        self._localObservableHandler.addTupleProvider(tupleName, provider=provider)

    def notifyOfTupleUpdate(self, tupleSelector: TupleSelector) -> None:
        """ Notify of Tuple Update

        Notifies the local observable that tuples have been updated

        """
        if not self._localObservableHandler.hasTupleProvider(tupleSelector.name):
            raise Exception("Local observable doesn't have tuple provider for %s"
                            " registered, Proxy is : %s" % (
                                tupleSelector.name, self._filt
                            ))

        self._localObservableHandler.notifyOfTupleUpdate(tupleSelector)

    ## ----- Implement proxy from here on in

    @inlineCallbacks
    def _process(self, payloadEnvelope: PayloadEnvelope, vortexUuid: str, vortexName: str,
                 sendResponse: SendVortexMsgResponseCallable, **kwargs):
        if vortexName == self._proxyToVortexName:
            yield self._processUpdateFromBackend(payloadEnvelope)

        else:
            yield self._processSubscribeFromFrontend(payloadEnvelope, vortexUuid,
                                                     sendResponse)

    def _processSubscribeFromFrontend(self, payloadEnvelope: PayloadEnvelope,
                                      vortexUuid: str,
                                      sendResponse: SendVortexMsgResponseCallable):
        tupleSelector: TupleSelector = payloadEnvelope.filt["tupleSelector"]

        # If the local observable provides this tuple, then use that instead
        if self._localObservableHandler.hasTupleProvider(tupleSelector.name):
            return self._localObservableHandler._process(payloadEnvelope=payloadEnvelope,
                                                         vortexUuid=vortexUuid,
                                                         sendResponse=sendResponse)

        # Add support for just getting data, no subscription.
        if payloadEnvelope.filt.get("unsubscribe", True):
            return self._handleUnsubscribe(tupleSelector, vortexUuid)

        elif payloadEnvelope.filt.get("subscribe", True) and self._subscriptionsEnabled:
            return self._handleSubscribe(payloadEnvelope, tupleSelector, sendResponse,
                                         vortexUuid)

        else:
            return self._handlePoll(payloadEnvelope, tupleSelector, sendResponse)

    def _handleUnsubscribe(self, tupleSelector: TupleSelector, vortexUuid: str):

        if not self._hasTupleSelector(tupleSelector):
            return

        cache = self._getCache(tupleSelector)
        try:
            cache.vortexUuids.remove(vortexUuid)
        except KeyError:
            pass

    def _handleSubscribe(self, payloadEnvelope: PayloadEnvelope,
                         tupleSelector: TupleSelector,
                         sendResponse: SendVortexMsgResponseCallable,
                         vortexUuid: str):

        # Add support for just getting data, no subscription.
        cache = self._getCache(tupleSelector)
        if cache and cache.lastServerPayloadDate is not None and cache.cacheEnabled:
            respPayloadEnvelope = PayloadEnvelope(filt=payloadEnvelope.filt,
                                                  encodedPayload=cache.encodedPayload,
                                                  date=cache.lastServerPayloadDate)

            d = respPayloadEnvelope.toVortexMsgDefer()
            d.addCallback(sendResponse)
            d.addErrback(vortexLogFailure, logger, consumeError=True)

        elif cache:
            self._sendRequestToServer(payloadEnvelope)

        else:
            cache = self._makeCache(tupleSelector)
            self._sendRequestToServer(payloadEnvelope)

        cache.vortexUuids.add(vortexUuid)
        # Allow the cache to be disabled
        cache.cacheEnabled = (
                cache.cacheEnabled
                and not payloadEnvelope.filt.get("disableCache", False)
        )

    def _sendRequestToServer(self, payloadEnvelope: PayloadEnvelope):
        payloadEnvelope.filt["observerName"] = self._observerName
        d = VortexFactory.sendVortexMsg(vortexMsgs=payloadEnvelope.toVortexMsg(),
                                        destVortexName=self._proxyToVortexName)
        d.addErrback(vortexLogFailure, logger, consumeError=True)

    def _sendUnsubscribeToServer(self, tupleSelector: TupleSelector):
        payloadEnvelope = PayloadEnvelope()
        payloadEnvelope.filt["tupleSelector"] = tupleSelector
        payloadEnvelope.filt["unsubscribe"] = True
        self._sendRequestToServer(payloadEnvelope)

    def _handlePoll(self, payloadEnvelope: PayloadEnvelope,
                    tupleSelector: TupleSelector,
                    sendResponse: SendVortexMsgResponseCallable):

        useCache = payloadEnvelope.filt.get('useCache', True)

        # Keep a copy of the incoming filt, in case they are using PayloadResponse
        responseFilt = copy(payloadEnvelope.filt)

        # Restore the original payload filt (PayloadResponse) and send it back
        def reply(payload):
            payload.filt = responseFilt
            d = payload.toVortexMsgDefer()
            d.addCallback(sendResponse)
            d.addErrback(vortexLogFailure, logger, consumeError=True)
            # logger.debug("Received response from observable")

        if useCache:
            cache = self._getCache(tupleSelector)
            if cache and cache.lastServerPayloadDate is not None and cache.cacheEnabled:
                payloadEnvelope.encodedPayload = cache.encodedPayload
                payloadEnvelope.date = cache.lastServerPayloadDate
                reply(payloadEnvelope)
                return

        # Track the response, log an error if it fails
        # 5 Seconds is long enough
        pr = PayloadResponse(
            payloadEnvelope,
            timeout=PayloadResponse.TIMEOUT - 5,  # 5 seconds less
            logTimeoutError=False
        )

        pr.addErrback(self._handlePrFailure, tupleSelector)
        pr.addErrback(vortexLogFailure, logger, consumeError=True)
        pr.addCallback(reply)

        self._sendRequestToServer(payloadEnvelope)

    def _handlePrFailure(self, f: Failure, tupleSelector):
        if f.check(TimeoutError):
            logger.error(
                "Received no response from\nobservable %s\ntuple selector %s",
                self._filt,
                tupleSelector.toJsonStr()
            )
        else:
            logger.error(
                "Unexpected error, %s\nobservable %s\ntuple selector %s",
                f,
                self._filt,
                tupleSelector.toJsonStr()
            )

    @deferToThreadWrapWithLogger(logger)
    def _processUpdateFromBackend(self, payloadEnvelope: PayloadEnvelope):

        tupleSelector: TupleSelector = payloadEnvelope.filt["tupleSelector"]

        if not self._hasTupleSelector(tupleSelector):
            return

        cache, requiredUpdate = self._updateCache(payloadEnvelope)
        if not requiredUpdate:
            return

        # Get / update the list of observing UUIDs
        observingUuids = cache.vortexUuids & set(VortexFactory.getRemoteVortexUuids())

        if not observingUuids:
            return

        # Create the vortexMsg
        vortexMsg = payloadEnvelope.toVortexMsg()

        # Send the vortex messages
        for vortexUuid in observingUuids:
            d = VortexFactory.sendVortexMsg(vortexMsgs=vortexMsg,
                                            destVortexUuid=vortexUuid)
            d.addErrback(vortexLogFailure, logger, consumeError=True)



from twisted.internet import defer
from twisted.trial import unittest

from vortex.Payload import Payload
from vortex.PayloadIO import PayloadIO
from vortex.PayloadResponse import PayloadResponse


class PayloadResponseTest(unittest.TestCase):
    def testSuccesCb(self):
        payload = Payload(filt={"key": "PayloadResponseTest"})

        payloadReponse = PayloadResponse(payload)
        payloadReponse.addCallback(lambda pl: self.assertIs(pl, payload))
        payloadReponse.addErrback(lambda pl: self.assertTrue(False))

        PayloadIO().process(payload, "", "", None, lambda _: None)

        return payloadReponse

    def testTimeoutErrback(self):
        payload = Payload(filt={"key": "PayloadResponseTest"})

        payloadReponse = PayloadResponse(payload, timeout=3)
        payloadReponse.addCallback(lambda pl: self.assertFalse(True))
        self.assertFailure(payloadReponse, defer.TimeoutError)

        # PayloadIO().process(payload)

        return payloadReponse

    def testSerialiseDeserialise(self):
        payload = Payload(filt={"key": "PayloadResponseTest"})

        d = PayloadResponse(payload, timeout=0)
        d.addErrback(lambda _: True)  # Handle the errback

        vortexMsg = payload.toVortexMsg()
        Payload().fromVortexMsg(vortexMsg)



"""
 *
 *  Copyright Synerty Pty Ltd 2013
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  Synerty Pty Ltd
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
 *
"""
import logging
from datetime import datetime
from typing import Dict
from typing import List
from typing import Union

from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

logger = logging.getLogger(__name__)


@addTupleType
class TestTuple(Tuple):
    """
    classdocs
    """

    __tupleType__ = "synerty.vortex.TestTuple"
    aInt: bool = TupleField()
    aFloat: float = TupleField()
    aString: str = TupleField()
    aStrWithUnicode: str = TupleField()
    aDate: datetime = TupleField()
    aBoolTrue: bool = TupleField()
    aBoolFalse: bool = TupleField()
    aSet: List[Union[str, int, bool]] = TupleField()
    aList: List[Union[str, int, bool, float]] = TupleField()
    aDict: Dict[Union[str, int], Union[str, float]] = TupleField()
    aSubTuple: "TestSubTuple" = TupleField()
    aListOfSubTuples: List["TestSubTuple"] = TupleField()

    def __init__(self, *args, **kwargs):
        # logger.debug("TestTuple constructed")
        Tuple.__init__(self, *args, **kwargs)


@addTupleType
class TestSubTuple(Tuple):
    """Test Sub Tuple
    This tuple will test if we can have tuples in tuples and still serialise and
    deserialise them.
    """

    __tupleType__ = "rapui.synerty.TestSubTuple"
    subInt: int = TupleField()



import logging

from twisted.internet import reactor

from vortex.VortexFactory import VortexFactory
from vortex.rpc.RPC import vortexRPC
from vortex.rpc.RPCTest import myRemoteAddMethod

logger = logging.getLogger(__name__)
logging.basicConfig()



if __name__ == '__main__':
    VortexFactory.createTcpServer("listenVortexName", 10101)
    reactor.callLater(0, logger.info, "RPCTest server running")
    reactor.callLater(0, myRemoteAddMethod.start)
    reactor.run()



import gzip
import logging
from datetime import datetime
from typing import List

from twisted.trial import unittest
from twisted.internet.error import TimeoutError as TwistedTimeoutError

from vortex.Tuple import Tuple, TupleField, addTupleType
from vortex.restful.GzippedDataHttpClient import GzippedDataHttpClient
from vortex.restful.RestfulResource import HTTP_REQUEST

logger = logging.getLogger(__name__)


@addTupleType
class GzippedDataHttpResponseTuple(Tuple):
    __tupleType__ = "GzippedDataHttpResponseTuple"

    requestDate: datetime = TupleField()
    responseDate: datetime = TupleField()
    code: int = TupleField(defaultValue=0)
    version: tuple = TupleField()  # (b'HTTP', 1, 1)
    headers: dict[str, str] = TupleField()
    body: bytes = TupleField()
    exceptions: List[Exception] = TupleField(defaultValue=[])


class GzippedDataHttpClientTest(unittest.TestCase):
    def setUp(self):
        self.payload = gzip.compress(b"test")
        self.headers = {}
        self.meta = GzippedDataHttpResponseTuple()

    def testHttpPost(self):
        client = GzippedDataHttpClient(
            url="http://httpbin.org/post",
            payload=self.payload,
            headers=self.headers,
            method=HTTP_REQUEST.POST,
            meta=GzippedDataHttpResponseTuple(),
            isPayloadGzipped=True,
            compressed=True,
        )
        d = client.run()
        d.addCallback(self._assertValidResponse)
        return d

    def testHttpsPost(self):
        client = GzippedDataHttpClient(
            url="https://httpbin.org/post",
            payload=self.payload,
            headers=self.headers,
            method=HTTP_REQUEST.POST,
            meta=GzippedDataHttpResponseTuple(),
            isPayloadGzipped=True,
            compressed=True,
        )
        d = client.run()
        d.addCallback(self._assertValidResponse)
        return d

    def testPlaintextDataPost(self):
        payload = gzip.decompress(self.payload)
        client = GzippedDataHttpClient(
            url="http://httpbin.org/post",
            payload=payload,
            headers=self.headers,
            method=HTTP_REQUEST.POST,
            meta=GzippedDataHttpResponseTuple(),
            isPayloadGzipped=False,
            compressed=True,
        )
        d = client.run()
        d.addCallback(self._assertValidResponse)
        return d

    def testConnectTimeout(self):
        client = GzippedDataHttpClient(
            url="http://example.com:19839",
            payload=self.payload,
            headers=self.headers,
            method=HTTP_REQUEST.POST,
            meta=GzippedDataHttpResponseTuple(),
            isPayloadGzipped=False,
            compressed=True,
            timeout=1.0,
        )
        d = client.run()
        d.addCallback(self._assertTimeoutException)
        return d

    def _assertValidResponse(self, tuple_: GzippedDataHttpResponseTuple):
        self.assertEqual(tuple_.code, 200)
        self.assertTrue(
            (tuple_.responseDate - tuple_.requestDate).total_seconds() > 0
        )

    def _assertTimeoutException(self, tuple_: GzippedDataHttpResponseTuple):
        booleans = [
            isinstance(e, TwistedTimeoutError) for e in tuple_.exceptions
        ]
        self.assertTrue(any(booleans))



import os
import shutil

from setuptools import setup, find_packages

package_name = "vortexpy"
package_version = '2.4.3'

egg_info = "%s.egg-info" % package_name
if os.path.isdir(egg_info):
    shutil.rmtree(egg_info)

requirements = [
    "SQLAlchemy >= 1.0.14",  # Database abstraction layer
    "GeoAlchemy2",  # Geospatial addons to SQLAlchemy
    "txwebsocket>=1.0.1",
    # txWS requires these, if we try to offline install the packages, txWS setup_requires
    # causes issues
    "vcversioner",
    "six",
    "ujson<2.0.0",
    "pytz",
    # RxPY by Microsoft. Used everywhere
    # TODO Upgrade to rx 3.x.x
    "rx < 3.0.0",
    "ddt >=1.4.1",
    # Test requirements
    "psutil",
]

setup(
    name="vortexpy",
    packages=find_packages(exclude=["test"]),
    # package_data={'vortex': ['*.xml']},
    version=package_version,
    install_requires=requirements,
    description="Synertys observable, routable, data serialisation and transport code.",
    author="Synerty",
    author_email="contact@synerty.com",
    url="https://github.com/Synerty/vortexpy",
    download_url=(
        "https://github.com/Synerty/%s/tarball/%s" % (package_name, package_version)
    ),
    keywords=["vortex", "observable", "http", "compressed", "synerty"],
    classifiers=[
        "Programming Language :: Python :: 3.5",
    ],
)



import logging
import sys
from copy import copy
from datetime import datetime
from typing import Optional
from uuid import uuid4

import pytz
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure

from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexFactory import VortexFactory

logger = logging.getLogger(__name__)


class PayloadResponse(Deferred):
    """ Payload Response

    This class is used to catch responses from a sent payload.
    If the remote end is going to send back a payload, with the same filt, this class can
    be used to capture this can and call it's call back.

    If the response is not received within the timeout, the errback is called.

    Here is some example usage.

    ::

        logger = logging.getLogger("Example Payload Response")
        payload = Payload(filt={"rapuiServerEcho":True})
        responseDeferred = PayloadResponse(payload)
        VortexFactory.sendVortexMsg(payload.toVortexMsg())
        logger.info("Payload Sent")

        responseDeferred.addCallbacks(logger.info, logger.error)

    """

    __messageIdKey = "PayloadResponse.messageId"

    PROCESSING = "Processing"
    # NO_ENDPOINT = "No Endpoint"
    FAILED = "Failed"
    SUCCESS = "Success"
    TIMED_OUT = "Timed Out"

    TIMEOUT: float = 30.00

    __SEQ = 1

    def __init__(self, payloadEnvelope: PayloadEnvelope,
                 destVortexName: Optional[str] = None,
                 destVortexUuid: Optional[str] = None,
                 timeout: Optional[float] = None,
                 resultCheck=True,
                 logTimeoutError=True) -> None:
        """ Constructor

        Tag and optionally send a payload.

        The timeout starts as soon as the constructor is called.

        :param payloadEnvelope The payloadEnvelope to send to the remote and, and wait for a response for
        :param destVortexName The name of the vortex to send to.
        :param destVortexUuid The UUID of the vortex to send a payload to.
        :param timeout The timeout to wait for a response
        :param resultCheck Should the response payload.result be checked, if it fails
                    it errback will be called.
        """
        Deferred.__init__(self)

        if not timeout:
            timeout = self.TIMEOUT

        self._resultCheck = resultCheck
        self._logTimeoutError = logTimeoutError

        # uuid4 can have collisions
        self._messageId = str(uuid4()) + str(PayloadResponse.__SEQ)
        PayloadResponse.__SEQ += 1

        payloadEnvelope.filt[self.__messageIdKey] = self._messageId
        self._filt = copy(payloadEnvelope.filt)
        self._destVortexName = destVortexName

        self._status = self.PROCESSING
        self._date = datetime.now(pytz.utc)

        self._endpoint = PayloadEndpoint(self._filt, self._process)

        if destVortexName or destVortexUuid:
            d: Deferred = payloadEnvelope.toVortexMsgDefer()
            d.addCallback(VortexFactory.sendVortexMsg,
                          destVortexName=destVortexName,
                          destVortexUuid=destVortexUuid)
            d.addErrback(self.errback)

        try:
            raise Exception()
        except:
            self._stack = sys.exc_info()[2]

        # noinspection PyTypeChecker
        self.addTimeout(timeout, reactor)
        self.addErrback(self._timedOut)

    @classmethod
    def isResponsePayloadEnvelope(cls, payloadEnvelope: PayloadEnvelope):
        """ Is Response Payload Envelope

        The PayloadResponse tags the payloads, so it expects a unique message back.

        :returns: True if this payload has been tagged by a PayloadResponse class
        """
        return cls.__messageIdKey in payloadEnvelope.filt

    @property
    def status(self):
        return self._status

    def _timedOut(self, failure: Failure):
        if self._endpoint:
            self._endpoint.shutdown()
            self._endpoint = None

        if self._logTimeoutError:
            logger.error("Timed out for payload %s", self._filt)
        self._status = self.TIMED_OUT
        return failure

    def _process(self, payloadEnvelope:PayloadEnvelope, vortexName, **kwargs):
        if self._endpoint:
            self._endpoint.shutdown()
            self._endpoint = None

        if self._destVortexName and vortexName != self._destVortexName:
            logger.debug("Received response from a vortex other than the dest vortex, "
                         "Expected %s, Received %s", self._destVortexName, vortexName)
            return

        if self.called:
            logger.error("Received response after timeout for %s" % self._filt)
            return

        if self._resultCheck and not payloadEnvelope.result in (None, True):
            self._status = self.FAILED
            self.errback(
                Failure(Exception(payloadEnvelope.result).with_traceback(self._stack)))

        else:
            self._status = self.SUCCESS
            self.callback(payloadEnvelope)

        # Return self, so that PayloadIO knows this is a asyncronous method
        return self



from datetime import datetime

import gc
import os
import psutil
import pytz

process = psutil.Process(os.getpid())

ITER_COUNT = 1000 * 1000 * 5

RESULT = None


def makeL(i):
    # Use this line to negate the effect of the strings on the test
    # return "Python is smart and will only create one string with this line"

    # Use this if you want to see the difference with 5 million unique strings
    return "This is a sample string %s" % i


def timeit(method):
    def timed(*args, **kw):
        global RESULT
        RESULT = None
        gc.collect()

        s = datetime.now(pytz.utc)
        startMem = process.memory_info().rss
        RESULT = method(*args, **kw)
        e = datetime.now(pytz.utc)
        endMem = process.memory_info().rss

        sizeMb = (endMem - startMem) / 1024 / 1024
        sizeMbStr = "{0:,}".format(round(sizeMb, 2))

        print('Time Taken = %s, \t%s, \tSize = %s' % (e - s, method.__name__, sizeMbStr))

    return timed


from vortex.Tuple import Tuple, addTupleType, TupleHash
from vortex.Payload import Payload


@addTupleType
class X(Tuple):
    __tupleType__ = "X"
    __slots__ = ["i", "l"]

    def __init__(self, i=None, l=None):
        self.i, self.l = i, l

@timeit
def provile_dict_of_nt():
    return [X(i=i, l=makeL(i)) for i in range(ITER_COUNT)]


if __name__ == "__main__":
    provile_dict_of_nt()

    tupleIn = X("val1", "val2")
    encodedPayload = Payload(tuples=[tupleIn]).toEncodedPayload()
    payload = Payload().fromEncodedPayload(encodedPayload)

    assert TupleHash(tupleIn) == TupleHash(payload.tuples[0])



from abc import ABCMeta, abstractmethod
from typing import Optional, Union, Callable, List

from twisted.internet.defer import Deferred

from vortex.PayloadEnvelope import VortexMsgList

SendVortexMsgResponseCallable = Callable[[Union[VortexMsgList, bytes], int], Deferred]


class VortexInfo:
    """ Vortex Info

    This class stores details about the end of a vortex.
    """

    def __init__(self, name, uuid):
        self.name, self.uuid = name, uuid

    def __repr__(self):
        return "VortexInfo, Name=%s, UUID=%s" % (self.name, self.uuid)


class VortexABC(metaclass=ABCMeta):
    @property
    @abstractmethod
    def localVortexInfo(self) -> VortexInfo:
        pass

    @property
    @abstractmethod
    def remoteVortexInfo(self) -> List[VortexInfo]:
        pass

    @abstractmethod
    def sendVortexMsg(self,
                      vortexMsgs: Union[VortexMsgList, bytes],
                      vortexUuid: Optional[str] = None):
        """ Send Vortex Msg

        Sends the vortex message to any conencted clients with vortexUuid.
        Or broadcast it to all connected vortex clients if it's None

        :param vortexMsgs: The vortex message(s) to send
        :param vortexUuid: The vortexUuid of the client to send to.
        """
